UNIT vpeaksr; {class = Apr 7 2002}

INTERFACE
USES Windows, Dialogs, SysUtils, vzlocom, varfun, vspectr, vpmodel;

CONST nmmax:INTEGER =214;   {maximum size of LS-matrix}
TYPE TPeakSearch = CLASS
  PUBLIC { Public declarations }
  kws:REAL;        {coefficient of compression}
  CONSTRUCTOR Create;
  PROCEDURE MarkTheIntervals(CONST sp:TSpectrum; k1,k2:INTEGER;
                            w0,w1,dwi:REAL);
  PROCEDURE DivideTheInterval(j1,j2:INTEGER);
  PROCEDURE MarkByEnergies(ar:ARRAY OF REAL; n,l1,l2:INTEGER);
  PROCEDURE FindPeriods(n:INTEGER; dper,per_min,sca:REAL);
  PRIVATE { Private declarations }
  ii1:INTEGER;     {current channel for width determ.}
  inp,jnp:INTEGER; {initial and final number of found peaks}
  asca:REAL;       {amplitude scale}
  kx,kn:INTEGER;
  r,ur,err,am,amx,amn:REAL;
  ff,fz,fzz:REAL;
  dw,dif1,dif2:REAL;
  nsip,kp,ip,ip1,ip2,i1,i2,nk,m,m1,mx:INTEGER;
  kse,ab,ass:REAL;
  PROCEDURE CheckForDoublet(i,ip,nk,m:INTEGER; VAR ok:BOOLEAN);
  PROCEDURE ZeroPeak(i,i1,i2:INTEGER; w:REAL);
  PROCEDURE FindPeakTops;
  PROCEDURE SensitivityAnalysis(w:REAL);
  PROCEDURE ResolutionAnalysis;{Resolution checks}
  PROCEDURE SearchHiddenPeaks;{Search hidden peaks}
  PROCEDURE CheckStatRes(w:REAL); {Check for statistical resolution}
  END;

VAR marka:ARRAY[1..1000] OF REAL;    {set of marked points}
    peria:ARRAY[1..1000] OF PERAR;   {set of marked points}
    mauxa:ARRAY[1..1000] OF MARKREC; {set of bounds of marked intervals}
    npeak:INTEGER;           {number of marked points}
    nper:INTEGER;            {number of periods}
    nint:INTEGER;            {number of marked intervals}
    kc:INTEGER;              {number of points of marked intervals}
    sens,res:REAL;           {Sensitivity, resolution while peaksearch}
    linx:INTEGER;            {Maximum length of the search interval}
    kwmerg:REAL;             {N widths while dividing into intervals}
    admerg:REAL;             {Constant addition to interval }
    cw0,cw1:REAL;     {width dependence on channel: w=cw1*ch+cw0}
    kw0,kw1:REAL;     {-"-}
    okwidth:BOOLEAN;  {is the width accurately given?}
    posa,dpos,dwi0,dwi1:ARRAY[0..4] OF REAL;
    i:INTEGER;
    deb:BOOLEAN;
    peaksr:TPeakSearch;
CONST pwdcomm:STRING ='Inexact peak width dependence;';

IMPLEMENTATION
PROCEDURE TPeakSearch.FindPeriods(n:INTEGER; dper,per_min,sca:REAL);
{look for the periodicities.
 ABS(Fourier), nk2 numbers in onl-array;
 sens,res - sensitivity and resolution of the search;
 per_min   = minimal level for periods to be detected;
 sca = number of compressed x's;
 peria =result str.: nper of [amplitude, position, width, per]
 dper  =  minimal distance between periods
 The period is computed by the formula
              Per = 2*nk2/(Pos-1)   (* sca)
 For.Version: 3111989; DELPHI: Nov 26 2004}
VAR   i,j,k,j1,j2,npf,ik:INTEGER;
      sum,back,sig,wi,ps,s,p,dif1,dif2,qc,ampl,qc1,ps1,ps2,del,amr:REAL;
LABEL L341,L342;
BEGIN res:=1; nk:=n-2; nper:=0; j:=1; amn:=onl[1]; sum:=0;
FOR i:=1 TO nk DO
 BEGIN amn:=AMIN1(amn,onl[i]); sum:=sum+onl[i];
 onl[i]:=onl[i]+(1E-10)*FLOAT(i);
 END;
sum:=(sum-onl[1]-onl[2])/FLOAT(nk-2); IF((amx-amn)=0)THEN Exit;
back:=amn;
               {M A I N   L O O P}
sig:=SQRT(onl[1]/FLOAT(nk)+back)*3; FOR i:=2 TO nk-1 DO
 BEGIN err:=sig*sens; {Check the channels 2 and 3}
 IF((i=2)OR(i=3))THEN
  BEGIN IF(((onl[i]-back)<err)OR(onl[i]<onl[i+1]))THEN CONTINUE;
  wi:=1; s:=onl[2]+onl[3]; ps:=onl[3]/s+2; p:=nk/(ps-1);
  {epr:=nk*(1/(ps-1)-1/(ps-1+wi));} GOTO L342;
  END;
 {Check the other channels}
 IF((onl[i-1]>onl[i])OR(onl[i+1]>onl[i]))THEN CONTINUE;
 dif1:=(onl[i+1]-onl[i-1])*0.5; dif2:=onl[i+1]-2*onl[i]+onl[i-1];
 qc:=dif2/SQRT(1+SQR(dif1)); dif1:=(onl[i+2]-onl[i-1])*0.5;
 ampl:=onl[i]+onl[i+1]; dif2:=onl[i+2]-ampl+onl[i-1];
 qc1:=dif2/SQRT(1+SQR(dif1)); qc:=AMIN1(qc,qc1);
 IF(qc>-3.0)THEN CONTINUE; IF((onl[i]-back)<err)THEN CONTINUE;
 {Find the area and the width;}
 GetPeakWidth(onl,i,nk,wi);
 IF(wi=-1)THEN wi:=2; IF(ABS(wi)>8)THEN CONTINUE; wi:=wi*0.5;
 j1:=MAX1(Float(i)-wi,1); j2:=MIN1(Float(i)+wi,nk-1); err:=sig*sens;
 IF(ABS(onl[i]-(onl[i-1]+onl[i+1])*0.5)<err)THEN CONTINUE;
 s:=0; ps:=0; FOR j:=j1 TO j2 DO
  BEGIN s:=s+AMAX1(onl[j]-back,0); ps:=ps+AMAX1(onl[j]-back,0)*(j-i);
  END;
 ps:=AMAX1(ps/AMAX1(1,s)+FLOAT(i),2);
 IF(FLOAT(nk)/(ps-1)<=per_min)THEN CONTINUE;
 L342:; Inc(nper); WITH peria[nper] DO
  BEGIN area:=s; pos:=ps; width:=wi; per:=Float(nk)*sca/(ps-1);
  END;
 END; {End of Main Loop}
{Check for the resolution;} IF(nper<=1)THEN Exit;
{Peak merging;} npf:=nper;
REPEAT amr:=1E+15; FOR i:=2 TO npf DO
 BEGIN ps1:=peria[i-1].pos-1; ps2:=peria[i].pos-1;
 del:=(ps2-ps1)*FLOAT(nk)*sca/(ps2*ps1);
 IF(del>amr)THEN CONTINUE; amr:=del; ik:=i;
 END;
IF(amr>=res*dper)THEN BREAK;
s:=peria[ik-1].area+peria[ik].area;
wi:=AMAX1(peria[ik-1].width,peria[ik].width);
ps:=(peria[ik-1].area*peria[ik-1].pos+peria[ik].area*peria[ik].pos)/s;
p:=FLOAT(nk)*sca/(ps-1); WITH peria[ik-1] DO
 BEGIN area:=s; pos:=ps; width:=wi; per:=p;
 END;
FOR i:=ik TO nper DO peria[i]:=peria[i+1]; Dec(npf);
UNTIL false;
nper:=npf;
END;
PROCEDURE TPeakSearch.MarkByEnergies(ar:ARRAY OF REAL; n,l1,l2:INTEGER);
{ add to automatically found peaks those of energy list ar,n items.
  Mar 20 2003}
VAR i,j,k:INTEGER;
BEGIN k:=0; FOR i:=1 TO n DO
 BEGIN r:=ar[i-1]; IF(infun(Round(r),l1,l2))THEN
  BEGIN marka[i+npeak]:=r; Inc(k);
  END;
 END;
npeak:=npeak+k; arran(marka,npeak,1); kp:=npeak-1;
IF(npeak>1)THEN
 BEGIN FOR i:=1 TO npeak-1 DO ha[i]:=marka[i+1]-marka[i];
 REPEAT FindMxn(ha,1,npeak-1,r,ur,ip1,m); err:=ABS(res);
 IF((ha[m]<err)OR(ha[m]<1.1))THEN
  BEGIN marka[m]:=(marka[m]+marka[m+1])*0.5; marka[m+1]:=0;
  cleanz(marka,npeak,0); arran(marka,npeak,1); IF(npeak<=1)THEN BREAK;
  FOR i:=1 TO npeak-1 DO ha[i]:=marka[i+1]-marka[i]; Dec(kp);
  END                        ELSE BREAK;
 UNTIL kp=0;
 DivideTheInterval(l1,l2); FOR i:=1 TO nint DO ha[i]:=0;
 FOR j:=1 TO n DO
  BEGIN r:=ar[j-1]; FOR i:=1 TO nint DO WITH mauxa[i] DO
   IF(infun(Round(r),i1,i2))THEN ha[i]:=1;
  END;
 FOR i:=1 TO nint DO WITH mauxa[i] DO
  IF(ha[i]=0)THEN FOR k:=k1 TO k2 DO marka[k]:=0;
 cleanz(marka,npeak,0);
 END;
END;
PROCEDURE TPeakSearch.FindPeakTops;
{find by quasi-curvatures and first error check}
VAR i,j:INTEGER; ok:BOOLEAN; r:REAL;
BEGIN FOR i:=2 TO m1 DO {compute quasi-curvs}
 BEGIN dif1:=(sa[i+1]-sa[i-1])*0.5; dif2:=sa[i-1]-2*sa[i]+sa[i+1];
 r:=1+dif1*dif1; ff:=dif2/SQRT(r);
 SLanz(m,i,r,dif1,dif2); r:=1+dif1*dif1; ff:=(ff+dif2/SQRT(r))/2;
 ff:=ff+i*0.000007; wa[i]:=ff;
 END;
nk:=0; FOR i:=2 TO m1 DO
 BEGIN ok:=((wa[i]<-1.5)AND(wa[i]<wa[i-1])AND(wa[i]<wa[i+1]));
 err:=sb[i]*sens*kse;
 ok:=ok AND (sa[i]>err)AND(sa[i+1]>err)AND(sa[i-1]>err);{error checks}
 IF(deb)THEN Writeln(fout,i+i1-1:6,' qc,b,e,sa ',wa[i]:7:2,sb[i]:7:2,
  err:7:2,sa[i]:7:0,ok:8);
 IF(ok)THEN
  BEGIN Inc(nk); j:=i; ja[nk]:=wa[i]; IF(sa[i+1]>sa[i])THEN Inc(j);
  IF(sa[i-1]>sa[i])THEN Dec(j); wa[nk]:=j;
  END;
 END;
IF(amx<500)THEN FOR i:=1 TO nk DO
 BEGIN ip1:=MAX0(1,Round(wa[i]-3)); ip2:=MIN0(m1,Round(wa[i]+3));
 ab:=0; ass:=0; FOR j:=ip1 TO ip2 DO
  BEGIN ab:=ab+ABS(sa[j+1]-sa[j]); ass:=ass+SQRT(sa[j]);
  END;
 IF(ab<(ass*sens*0.33))THEN
  BEGIN wa[i]:=0; IF(deb)THEN Writeln(fout,'dS < Serr',ab:9:0,ass:9:0);
  END;
 END;
cleanz(wa,nk,0);
FOR i:=1 TO nk-1 DO
 BEGIN ip1:=Round(wa[i]); ip2:=Round(wa[i+1]);
 ab:=sa[ip1]; ass:=sa[ip2];
 IF((AMAX1(ab,ass)<150)AND(Float(ip2-ip1)<AMAX1(dw*kws,2)))THEN
  BEGIN wa[i+1]:=(ab*ip1+ass*ip2)/AMAX1(1,ab+ass); wa[i]:=0;
  END;
 cleanz(wa,nk,0);
 END;
END;
PROCEDURE TPeakSearch.MarkTheIntervals(CONST sp:TSpectrum;
                      k1,k2:INTEGER; w0,w1,dwi:REAL);
{Find the peaks or multiplets;  version 14.10.84; DELPHI=Feb 16 2002}
VAR   i,j,k,jc1,jc2,linxc:INTEGER; nok:BOOLEAN;
      a,p,x,z,rl,w:REAL;
LABEL L1,L200,Lend;
 PROCEDURE MakeEnvelope(i1,i2:INTEGER);
 VAR i:INTEGER;
 BEGIN a:=(sa[i2]-sa[i1])/Float(i2-i1); p:=sa[i1]-a;
 FOR i:=i1 TO i2 DO sd[i]:=a*i+p;
 END;
{---------------------------------------------------------------}
BEGIN npeak:=0; jc1:=MAX0(2,k1); jc2:=k2-1; okwidth:=true; dw:=dwi;
cw1:=(w1-w0)/AMAX0(1,jc2-jc1); cw0:=w1-cw1*jc2;

{******} deb:=false; IF(debugV.activ)THEN deb:=true;  {******}

IF(deb)THEN
 BEGIN Writeln(fout,'kw1 kw0 for width ',cw1:9:4,cw0:9:3);
 END;
IF((jc2-jc1)<=Round((w0+w1)/2))THEN GOTO Lend; i1:=jc1;
kws:=1; IF(w0>8)THEN kws:=w0/8; linxc:=Round(Float(linx)*kws);
        {Analyze current interval}
L1: IF((jc2-i1+1)<linxc)THEN i2:=jc2 ELSE i2:=MIN0(i1+linxc-1,jc2);
IF((jc2-i1)<=Round((w0+w1)/2))THEN GOTO Lend;
kw1:=cw1; kw0:=cw0;  FOR j:=i1 TO i2 DO
 BEGIN k:=j-i1+1; z:=sp.y[j]+j*0.000001; sa[k]:=AMAX1(z,0);
 END;
m:=i2-i1+1; m1:=MIN0(1+(linxc DIV 3)*2,m); IF(m1<m)THEN
 BEGIN FindMxn(sa,m1,m,amx,amn,kx,kn); i2:=i2-m+kn;
 END;
        {Compress if needed}
m:=i2-i1+1; mx:=m; j:=m; IF(kws<>1)THEN
 BEGIN CompressArray(j,kws,m); sd[1]:=sd[2]; sd[m]:=sd[m-1];
 FOR i:=1 TO m DO sa[i]:=sd[i]; kw0:=kw0/kws; kw1:=kw1/kws;
 END;
{take account of the small statistics}
IF(deb)THEN Writeln(fout,'                Interval = ',i1:6,i2:6);
rl:=1; m1:=m-1; FindMxn(sa,1,m,amx,amn,kx,kn); kx:=MAX0(2,MIN0(kx,m1));
x:=ABS(sa[kx+1]+sa[kx]+sa[kx-1]-3*amn);
IF(x<(3*SQRT(ABS(3*amn+1))*sens))THEN
 BEGIN i1:=i2; GOTO L1;
 END;
IF(amx<=150)THEN
 BEGIN FOR i:=1 TO m1 DO sa[i]:=(sa[i]+sa[i+1]);
 sa[m]:=sa[m]+sa[m1]; amx:=AMAX1(sa[kx-1]+amx,amx+sa[kx+1]);
 END;
{smooth the spectrum}
x:=AMAX1(sa[1],sa[m]); kse:=1-ABS(x)/AMAX1(1,ABS(x)+ABS(x));
kse:=kse*(4*SQRT(EXP(-SQR((amx-500)/2500)))+2); kse:=AMIN1(kse,3);
asca:=amx;
IF(deb)THEN Writeln(fout,'kse= ',kse:9:3,'  max,n= ',asca:9:0,amn:8:0);
z:=10000/AMAX1(amx,1); IF(amx<=10000)THEN z:=1; amx:=amx*z;
FOR i:=1 TO m DO sa[i]:=z*sa[i];
FOR i:=1 TO m DO sb[i]:=SQRT(ABS(sa[i])+1);{errors}
AutoSmooth(m,1,0.5,nok);
FOR i:=2 TO m1 DO sa[i]:=AMAX1(sd[i],0);
{construct the background spline}
IF(amx>150)THEN
 BEGIN AutoEnvelope(m,7); {GetEnvelope(m,7,0.5,0.5,0.5);}
 END       ELSE
 BEGIN IF(NOT((kn=1)OR(kn=m)))THEN
  BEGIN MakeEnvelope(1,kn); MakeEnvelope(kn,m);
  END                         ELSE MakeEnvelope(1,m);
 END;
IF((deb)AND(amx>150))THEN Writeln(fout,'envel.lambd= ',qq.q1:12:8);
FOR i:=1 TO m DO sa[i]:=AMAX1(sa[i]-sd[i],0);{subtract background}
FindMxn(sa,1,m,amx,amn,kx,kn);
{IF(deb)THEN FOR i:=1 TO m DO onl[i]:=sa[i];}

                  {MAIN}
FindPeakTops; w:=kw1*(i2-i1)/2+kw0;
IF(nk=0)THEN GOTO L200; SensitivityAnalysis(w);
IF(amx<100)THEN GOTO L200;
ResolutionAnalysis;
IF(nk=0)THEN GOTO L200; SearchHiddenPeaks;
CheckStatRes(w);
                  {FINISH}
L200: m:=Round(m*kws);
FOR i:=1 TO nk DO wa[i]:=wa[i]*kws+i1-1;   {Decompress: to real X-scale}
IF(deb)THEN FOR i:=1 TO nk DO Writeln(fout,wa[i]:12:2);
 {WITH LB1rec DO
  BEGIN nitems:=3;
  ar[1]:=IntToStr(i1)+'  '+IntToStr(i2);
  ar[2]:=IntToStr(nk); ar[3]:=IntToStr(npeak);
  END;
 MakeForm(0);
 GraphSpecAndModel(Form2.Image1,m,1,nk); Form2.ShowModal;
 }
                   {Loop end}
k:=((nmmax-14) DIV 2); IF((npeak+nk)<k)THEN
 BEGIN FOR i:=1 TO nk DO marka[npeak+i]:=ABS(wa[i]);
 npeak:=npeak+nk; i1:=i2; GOTO L1; {loop end}
 END;
ShowMess(IntToStr(k)+' peaks already found, search stopped');
nk:=MIN0(k-npeak,nk);
FOR i:=1 TO nk DO marka[npeak+i]:=ABS(wa[i]); npeak:=k;
Lend: ;
END;
PROCEDURE TPeakSearch.DivideTheInterval(j1,j2:INTEGER);
VAR i,j,k,ii,i2n,k2n:INTEGER; ok:BOOLEAN; rr,a,ac0,qc0,w:REAL;
BEGIN nint:=0; IF(npeak=0)THEN Exit; arran(marka,npeak,1);
WITH pmod[1] DO {construct the single peak}
 BEGIN j:=2*nl+1; a:=1/morr[nl+1]; w:=cw1*(j1+j2)*0.5+cw0;
 FOR i:=1 TO j DO
  BEGIN Getffzfzz((Float(i)-Float(nl+1))/w,ff,fz,fzz); ha[i]:=a*ff;
  END;
 {determine the asimmetry and curvature}
 ac0:=0; a:=morr[nl+1]; FOR i:=1 TO j DO
  BEGIN r:=(i-nl-1)/w; IF(ha[i]>(a/32))THEN ac0:=ac0+r*ha[i]/ha[nl+1];
  END;
 i:=nl+1; dif1:=(ha[i+1]-ha[i-1])*0.5;
 dif2:=ha[i-1]-2*ha[i]+ha[i+1]; dif1:=dif1*amx; dif2:=dif2*amx;
 r:=1+dif1*dif1; qc0:=dif2/SQRT(r);
 END;
{AssignFile(fout,'\zlok\a.sav'); Rewrite(fout);
FOR i:=1 TO 10 DO
 BEGIN FOR j:=1 TO 10 DO Write(fout,marka[j+(i-1)*10]:7:0);
 Writeln(fout);
 END;}
wa[1]:=j1; FOR i:=1 TO npeak DO wa[i+1]:=marka[i]; wa[npeak+2]:=j2;
r:=kwmerg; FOR i:=1 TO npeak+1 DO
 BEGIN w:=(kw1*wa[i+1]+kw0)*kws; j:=(i-1)*3+1;
 ja[j]:=-AMAX1(wa[i],wa[i+1]-r*w-admerg); ja[j+1]:=wa[i+1];
 ja[j+2]:=-AMIN1(wa[i+2],wa[i+1]+r*w+admerg);
 IF(deb)THEN Writeln(fout,kw1:9:4,kw0:9:4,w:9:3,wa[i+1]:9:3);
 END;
FOR i:=1 TO npeak DO
 BEGIN k:=(i-1)*3; ip1:=Round(ABS(ja[k+3])); ip2:=Round(ABS(ja[k+4]));
 FindMxn(onl,ip1-1,ip1+1,amx,r,kx,kn); ja[k+3]:=-kn;
 FindMxn(onl,ip2-1,ip2+1,amx,r,kx,kn); ja[k+4]:=-kn;
 END;
FOR i:=1 TO npeak DO
 BEGIN k:=(i-1)*3; IF(ABS(ja[k+3])>ABS(ja[k+4]))THEN
  BEGIN ja[k+3]:=0; ja[k+4]:=0;
  END;
 END;
w:=(kw1*marka[1]+kw0)*kws;
ja[1]:=-MAX1(Float(j1),marka[1]-kwmerg*w-admerg);
IF(deb)THEN Writeln(fout,w:9:3);
w:=(kw1*marka[npeak]+kw0)*kws;
IF(deb)THEN Writeln(fout,w:9:3);
ja[npeak*3]:=-MIN1(Float(j2),marka[npeak]+kwmerg*w+admerg);
ii:=3*npeak; ok:=true; k:=0; FOR i:=1 TO ii DO
 BEGIN j:=sign(ja[i]);
 CASE j OF
 -1: BEGIN IF(ok)THEN
      BEGIN Inc(nint); mauxa[nint].i1:=Round(ABS(ja[i]));
      mauxa[nint].k1:=k+1;
      END        ELSE
       BEGIN mauxa[nint].i2:=Round(ABS(ja[i])); mauxa[nint].k2:=k;
       END;
     ok:=NOT ok;
     END;
  1: Inc(k);
 ELSE END;
 END;
{Form small intervals}
REPEAT j:=0; FOR i:=1 TO nint DO WITH mauxa[i] DO
 IF((i2-i1+1)>j)THEN
  BEGIN j:=i2-i1+1; ii:=i;
  END;
IF(j<=linx)THEN BREAK; WITH mauxa[ii] DO
 BEGIN k:=0; FOR i:=k1 TO k2-1 DO IF((marka[i+1]-marka[i]+1)>k)THEN
  BEGIN k:=Round(marka[i+1]-marka[i]+1); j:=i;
  END;
 i2n:=mauxa[ii].i2; k2n:=mauxa[ii].k2;
 FOR i:=nint+1 DOWNTO ii+1 DO mauxa[i]:=mauxa[i-1];
 i:=Round(Float(k)/2); Inc(nint);
 {IF(deb)THEN Writeln(fout,i:5,j:5,marka[j]:10:1);}
 mauxa[ii].i2:=Round(marka[j])+i;
 mauxa[ii].k2:=j;

 mauxa[ii+1].i1:=Round(marka[j])+i;
 mauxa[ii+1].i2:=i2n;
 mauxa[ii+1].k2:=k2n;
 mauxa[ii+1].k1:=j+1;
 END;
{IF(deb)THEN FOR i:=1 TO nint DO
 WITH mauxa[i] DO Writeln(fout,i1:5,i2:5,k1:4,k2:4);}
UNTIL false;
{IF(deb)THEN
 BEGIN FOR i:=1 TO nint DO WITH mauxa[i] DO
  Writeln(fout,'Intervals ',i:6,i1:5,i2:5,k1:5,k2:5);
 END;}
END;
PROCEDURE TPeakSearch.SensitivityAnalysis(w:REAL);
{Check by sensitivity criteria}
VAR i,j,k:INTEGER;
BEGIN wa[nk+1]:=m; IF(deb)THEN Writeln(fout,
 '         pos     qc      ampl   err     area     err     spu= ');
FOR i:=1 TO nk DO ha[i+1]:=wa[i+1]-wa[i]; ha[1]:=wa[1]-1; {differences}
FOR i:=1 TO nk DO
 BEGIN ab:=0; ass:=0; k:=Round(wa[i]);
 ip1:=MAX0(1,k-MIN1(w,ha[i]/2)); ip2:=MIN0(m,k+MIN1(w,ha[i+1]/2));
 FOR j:=ip1 TO ip2 DO
  BEGIN ass:=ass+sa[j]; ab:=ab+SQR(sb[j]);
  END;
 ok:=true; err:=SQRT(ab)*sens*3*kse; IF(ass<err)THEN wa[i]:=0;
 IF(deb)THEN
  Writeln(fout,'SensA',(k+i1-1):8,
   ja[i]:7:2,sa[k]:8:0,sb[k]:7:0,ass:9:2,err:9:2,' spu= ',ok);
 END;
cleanz(wa,nk,0); FOR i:=1 TO m DO {Zero weak counts}
 IF(NOT(sa[i]>=sb[i]*sens*kse))THEN sa[i]:=0;
END;
PROCEDURE TPeakSearch.ResolutionAnalysis;{Resolution checks}
VAR   i,j,k,n:INTEGER; ok,ok1,ok2,ok3:BOOLEAN; w,a,wc,wr,wm:REAL;
 PROCEDURE FalseOK;
 BEGIN ok1:=false; ok2:=false; ok3:=false;
 END;
LABEL L1;
BEGIN n:=0; a:=0; ab:=0; IF(nk<=1)THEN GOTO L1; FOR i:=1 TO nk DO
 BEGIN wc:=kw1*(wa[i]+i1-1)+kw0; ip:=Round(wa[i]);
 GetPeakWidth(sa,ip,m,wr);
 ha[i]:=wr; IF(wr<>-1)THEN a:=a+wr ELSE a:=a+wc; ab:=ab+wc;
 END;
GetMedian(ha,nk,wm); wr:=a/AMAX0(1,nk); wc:=ab/AMAX0(1,nk);
IF((ABS(wr-wm)>3*dw*kws)OR(ABS(wr-wc)>3*dw*kws))THEN okwidth:=false;
IF(deb)THEN Writeln(fout,'WidthA      real= ',wr:7:2,'  median= ',
 wm:7:2,'  calib = ',wc:7:2,'  dw= ',dw*kws:6:2);
{Purge: weak + broad peaks + not isolated;   ha=array of real widths}
IF(deb)THEN
 Writeln(fout,'ResA ','     pos     counts   error  w_calib w_found');
FOR i:=1 TO nk DO
 BEGIN wc:=kw1*(wa[i]+i1-1)+kw0; ip:=Round(wa[i]);
 err:=3*ABS(sb[ip])*sens; IF(sa[ip]>=err)THEN ok1:=true ELSE ok1:=false;
 IF((ABS(ha[i]-wc)<=dw*kws*ABS(res))AND(ha[i]<>-1))THEN ok2:=true
                                                   ELSE ok2:=false;
 k:=i; IF(i=nk)THEN k:=-1;
 CASE k OF
 1:  IF((wa[i+1]-wa[i])>(2*wc))THEN ok3:=true ELSE ok3:=false;
 -1: IF((wa[i]-wa[i-1])>(2*wc))THEN ok3:=true ELSE ok3:=false;
 ELSE IF(((wa[i+1]-wa[i])>(2*wc))AND((wa[i]-wa[i-1])>(2*wc)))THEN
      ok3:=true ELSE ok3:=false;
      END;
 IF(ha[i]=-1)THEN ok3:=false;  {peak without width is weak}
 IF(rinfun(ha[i],0.5,2))THEN FalseOK;  {too narrow peak}
 IF((sa[ip]<err)AND(ha[i]<wc))THEN FalseOK;  {false peak}
 IF(sa[ip]<err*0.5)THEN FalseOK;  {false peak}
 IF((NOT ok1)AND(NOT ok2)AND(sa[ip]<=200))THEN FalseOK;  {false peak}
 IF(deb)THEN Writeln(fout,'    ',(ip+i1-1):9,sa[ip]:9:0,err:9:0,wc:9:2,
  ha[i]:9:2,'   ',ok1:6,' ',ok2:6,' ',ok3:6);

 ok:=((ha[i]-wc)>(dw*1.5*kws*ABS(res)))AND((dw/wc)<0.25);
 IF(ok1 AND(NOT ok2)AND(ok3)AND(ok)AND(sa[ip]>200))THEN
  BEGIN CheckForDoublet(i,ip,nk,m,ok); IF(ok)THEN  {large broad}
   BEGIN Inc(n); j:=MAX0(Round(wc/2),3); wa[i]:=ip-j; wa[nk+n]:=ip+j;
   END;
  IF(deb)THEN
   BEGIN s:='resolved'; IF(ok)THEN s:='un'+s; Write(fout,s,ip+i1-1:9);
   IF(ok)THEN
    Writeln(fout,' replaced by ',wa[i]+i1-1:9:2,' ',wa[nk+n]+i1-1:9:2);
   END;
  END;
 j:=0; IF(ok1)THEN j:=j+4; IF(ok2)THEN j:=j+2; IF(ok3)THEN j:=j+1;
 ja[i]:=j;
 END;
FOR i:=1 TO nk DO
 BEGIN IF(Round(ja[i])=0{<4})THEN wa[i]:=0;
 k:=Round(wa[i]); IF((k<>0)AND(asca<20))THEN  {check for spurious}
  BEGIN ip1:=MAX0(1,k-MIN1(w,wm/2)); ip2:=MIN0(m,k+MIN1(w,ha[i+1]/2));
  a:=sa[k]; kp:=0;
  FOR j:=ip1 TO ip2 DO IF(ABS(sa[j]-a)>(sb[j]*sens*kse))THEN Inc(kp);
  IF((kp<3)AND((Round(ja[i])AND(2)=0))AND(ip2-ip1>=5))THEN
   BEGIN wa[i]:=0; ok:=false; Writeln(fout,'spur ',k+i1-1:5,kp:4);
   END;
  END;
 END;
nk:=nk+n; cleanz(wa,nk,0); arran(wa,nk,1); {Add the splitted doublets}

L1: FOR i:=1 TO nk DO
 BEGIN wc:=kw1*(wa[i]+i1-1)+kw0; ZeroPeak(i,1,m,wc);  {Purge peaks}
 END;
FOR i:=2 TO m-1 DO                     {Purge single and double counts}
 IF((sa[i-1]=0)AND(sa[i]>0)AND(sa[i+1]=0))THEN sa[i]:=0;
FOR i:=3 TO m-1 DO
 IF((sa[i-2]=0)AND(sa[i-1]>0)AND(sa[i]>0)AND(sa[i+1]=0))THEN
  BEGIN sa[i-1]:=0; sa[i]:=0;
  END;
END;
PROCEDURE TPeakSearch.SearchHiddenPeaks;{Search hidden peaks}
VAR i,j,k,n,iw:INTEGER; w,z,a:REAL;
BEGIN w:=kw1*m*0.5+kw0; iw:=Round(w); n:=0;
REPEAT FindMxn(sa,1,m,amx,z,ip,j); IF(amx<16)THEN BREAK;
FOR i:=ip DOWNTO MAX0(1,ip-iw) DO IF(sa[i]=0)THEN BREAK;
ip1:=i; FOR i:=ip TO MIN0(m,ip+iw) DO IF(sa[i]=0)THEN BREAK; ip2:=i;
ass:=0; err:=0; k:=0; FOR i:=ip1 TO ip2 DO
 BEGIN ass:=ass+sa[i]; err:=err+2*ABS(sb[ip])*sens*kse;
 IF(sa[i]>0)THEN Inc(k);
 END;
w:=kw1*ip+kw0+0.5; WITH pmod[1] DO
 BEGIN a:=sa[ip]/morr[nl+1]; FOR i:=1 TO m DO
  BEGIN z:=(i-ip)/w; Getffzfzz(z,ff,fz,fzz); sa[i]:=AMAX1(sa[i]-a*ff,0);
  END;
 END;
IF((ass>err)AND((k+1)>=iw))THEN
 BEGIN Inc(n); wa[nk+n]:=ip;
 IF(deb)THEN Writeln(fout,'Hidda ip,ass,e,k,iw ',
  ip+i1-1:5,ass:9:0,err:9:0,k:5,iw:5,'    Added ',ip+i1-1:5);
 END;
UNTIL false;
nk:=nk+n; arran(wa,nk,1); {Add the hidden peaks}
END;
PROCEDURE TPeakSearch.CheckStatRes(w:REAL);
{Check for statistical resolution}
VAR i,j,k:INTEGER;
BEGIN j:=nk-1; IF(nk>1)THEN
 BEGIN FOR i:=1 TO nk-1 DO ha[i]:=wa[i+1]-wa[i];
 REPEAT FindMxn(ha,1,nk-1,r,ur,ip,j);
 k:=Round(wa[j]); ip:=Round(wa[j+1]);
 err:=2*w*ABS(1/sb[k]+1/sb[ip])*ABS(res); err:=AMAX1(err,1);
 IF((ha[j]<err)OR((ha[j]<2)AND(AMIN1(sb[k],sb[ip])<200)))THEN
  BEGIN wa[j]:=(wa[j]+wa[j+1])*0.5; wa[j+1]:=0;
  IF(deb)THEN Writeln(fout,
   'ChesrA  Merged ',k:5,ip:5,' to ',wa[j]+i1-1:9:1,' err= ',err:9:2);
  cleanz(wa,nk,0); arran(wa,nk,1); IF(nk<=1)THEN BREAK;
  FOR i:=1 TO nk-1 DO ha[i]:=wa[i+1]-wa[i]; Dec(j);
  END                                                    ELSE BREAK;
 UNTIL j=0;
 END;
END;
PROCEDURE TPeakSearch.CheckForDoublet(i,ip,nk,m:INTEGER;VAR ok:BOOLEAN);
{check whether the real large width covers a doublet}
VAR ii,iw:INTEGER;
BEGIN ip1:=i-1; IF(i=1)THEN ip1:=i; ip2:=i+1; IF(i=nk)THEN ip2:=nk;
iw:=Round(ha[i]); FOR ii:=ip1 TO ip2 DO iw:=MAX0(iw,Round(ha[ii]));
IF(i=1)THEN ip1:=1 ELSE ip1:=Round(wa[i-1]);
IF(i=nk)THEN ip2:=m ELSE ip2:=Round(wa[i+1]);
ip1:=ip-iw-ip1; ip2:=ip2-ip-iw;
ok:=false; IF((ip1>0)AND(ip2>0))THEN ok:=true;{if true, unres.doublet}
END;
PROCEDURE TPeakSearch.ZeroPeak(i,i1,i2:INTEGER; w:REAL);
{Build the logical peak function p(x), and multiply by 1-p(x)}
VAR ii,ip:INTEGER; a,r,p,z,fl,fr:REAL;
BEGIN r:=1/AMAX1(1,w); p:=wa[i]; ip:=Round(p);
IF(sa[ip-1]>sa[ip])THEN ip:=ip-1 ELSE
IF(sa[ip+1]>sa[ip])THEN ip:=ip+1 ELSE; p:=ip;
WITH pmod[1] DO
 BEGIN a:=sa[ip]/morr[nl+1]; FOR ii:=i1 TO i2 DO
  BEGIN z:=(Float(ii)-p)/w;
  calf(z,ff,fz,fzz); calf(z+r,fr,fz,fzz); calf(z-r,fl,fz,fzz);
  sa[ii]:=AMAX1(sa[ii]-a*AMAX3(ff,fr,fl),0);
  END;
 END;
END;
CONSTRUCTOR TPeakSearch.Create;
BEGIN npeak:=0; nint:=0; kc:=0; sens:=1; res:=1; linx:=128;
kwmerg:=2; admerg:=3;
END;
{----------}
BEGIN FOR i:=1 TO 4 DO
 BEGIN dpos[i]:=5; dwi0[i]:=4; dwi1[i]:=100;
 END;
END.
