UNIT vspectr; {Jan 2 2002,  class = Feb 25 2002}

INTERFACE
USES Windows, Controls, Dialogs, SysUtils, vzlocom, varfun;

CONST nspch:INTEGER = 16384;{max number of channels in spectrum}
      DBackRec:LREC = (ar:('  No background','Constant','Linear',
        'Quadratic','Cubic','Of 4th order','Of 5th order',
        '  Real shape','','','','',''); nitems:8; indx:4);
      SpNumRec:LREC = (ar:(' 0 (effective)',' 1',' 2',' 3',
        ' 4','','','','','','','',''); nitems:5; indx:2;);
      SetBckg:LREC = (ar:(' Polynomial',' Real background',
      '','','','','','','','','','',''); nitems:2; indx:1;);

TYPE TSpectrum = CLASS
  PUBLIC { Public declarations }
  sfn,efn:STRING;      {sp & var.sp. names}
  smosp,emosp:INTEGER; {sp & var.sp. types}
  a1,a2,a11,a22:INTEGER; {spectrum & interval bounds}
  spx,spn:REAL;  {spectrum max & min}
  tofang:INTEGER;{tof(0) / angle distr.(1) / any file(2)}
  nconst:REAL;   {normalizing constant}
  abs1,abs2:REAL;{for absorption}
  shape:INTEGER; {sample form,0=cylinder,1=plane,2=sphere}
  spin:BOOLEAN;  {has been sp. read in?}
  effn:BOOLEAN;  {Normalized by eff.sp.?}
  absn:BOOLEAN;  {Corrected for absorption?}
  dback:INTEGER; {Mode and degree of background}
  bkpar:ARRAY[1..6] OF REAL; {background parameters}
  acor:REAL;       {exponential multiple}
  chw,delay,theta,FPath,w0,dw0,w1,dw1,lambda:REAL; {ToF & width info}
  wv,dwv:REAL;     {variable width for gamma-Ray analysis}
  x1,x2,delx,celx0,celx1:REAL; {for forming homogenious variables}
  y:ARRAY[1..16385] OF REAL; {spectrum}
  e:ARRAY[1..16385] OF REAL; {error}
  x:ARRAY[1..16385] OF REAL; {for fit or y-values}
  CONSTRUCTOR Create(vmdir:STRING);
  FUNCTION chtod(ch:REAL):REAL;
  FUNCTION dtoch(d:REAL):REAL;
  PROCEDURE Write2Binary(i1,i2:INTEGER; s:STRING);
  PROCEDURE Write4Binary(i1,i2:INTEGER; s:STRING);
  PROCEDURE WriteTxt(i1,i2:INTEGER; s:STRING; nd:INTEGER;
            sc:BOOLEAN);
  PROCEDURE WriteTxt3(i1,i2:INTEGER; ss:STRING; nd:INTEGER;
            sc,chang:BOOLEAN);
  PROCEDURE Read2Binary(VAR ok:BOOLEAN);
  PROCEDURE Read4Binary(VAR ok:BOOLEAN);
  PROCEDURE ReadTxt(k1,k2:INTEGER; VAR ok:BOOLEAN);
  PROCEDURE ReadTxt3(k1,k2:INTEGER; VAR ok:BOOLEAN);
  PROCEDURE Vancal(np:INTEGER; tau,Fp,zd:REAL);
  PROCEDURE fismax1;
  FUNCTION NoSpectrum(ns:INTEGER):BOOLEAN;
  PROCEDURE polcc(n1,n11,n12,n2:INTEGER; VAR ae,be,x0:REAL);
  PROCEDURE corbc(l,ll:INTEGER; VAR nbl,nz:INTEGER);
  PROCEDURE ReverseBytes(i:INTEGER; VAR o:INTEGER);
  PROCEDURE CalculateBack(VAR j1,j2,l:INTEGER);
  PROCEDURE CheckForSingleZeroes;
  PROCEDURE CheckErrorsFor0;
  PRIVATE { Private declarations }
  END;


VAR
    spvar:INTEGER;              {key: 0=spectrum, 1=variance}
    vartype:INTEGER;            {key: 0=error, 1=variance, 2=weight}
    errset:BOOLEAN;             {Are errors set?}
    smast:REAL;                 {small statistics level}
    covar:REAL;                 {Variance=.9*sp+.1*(covar*sp)^2}
    order,chibc:REAL;           {For blunder correction}
    skiph:BOOLEAN;              {Skip the spectrum 1024 bytes header}
    revby:BOOLEAN;              {Reverse bytes in 4-bytes words}
    chan1:BOOLEAN;              {Starting channel number = 1}
    i,j:INTEGER;

IMPLEMENTATION
CONSTRUCTOR TSpectrum.Create(vmdir:STRING);
VAR j:INTEGER;
BEGIN a1:=1; a2:=8192; a11:=a1; a22:=a2; spx:=1; spn:=0;
sfn:=vmdir; efn:=vmdir; smosp:=2; emosp:=2; vartype:=0;
FPath:=25; chw:=64; delay:=0; theta:=75; lambda:=1.54;
w0:=5; w1:=0; dw0:=4; dw1:=10; dback:=3; wv:=0; dwv:=0;
tofang:=0; abs1:=0.00147; abs2:=0.00893; shape:=0; nconst:=1; acor:=0;
spin:=true; effn:=false; absn:=false; errset:=false;
FOR j:=1 TO 6 DO bkpar[j]:=0;
FOR j:=1 TO 16384 DO
 BEGIN y[j]:=0; e[j]:=1; x[j]:=j;
 END;
x1:=1; x2:=16384; delx:=1; celx1:=1; celx0:=0;
END;
PROCEDURE TSpectrum.Read2Binary(VAR ok:BOOLEAN);
{Purpose: To read 2-bytes per word binary files, Version: Mar 30 2000}
VAR   res,ndel,i,j,k,nb:INTEGER; fb:FILE OF WORD; r:REAL;
BEGIN ndel:=256; ok:=false; s:=sfn; IF(spvar=1)THEN s:=efn; nb:=0;
AssignFile(fb,s); {$I-} Reset(fb); {$I+}; IF(IOresult<>0)THEN
 BEGIN ShowMess(s+': opening failure; '+crc+
 'file may be Read_Only'); Exit;
 END;
IF(skiph)THEN
 BEGIN BlockRead(fb,buf2,ndel,res); nb:=nb+res;
 BlockRead(fb,buf2,ndel,res); nb:=nb+res;
 END;
j:=1; WHILE(NOT Eof(fb))DO
 BEGIN BlockRead(fb,buf2,ndel,res); FOR i:=1 TO res DO
  BEGIN k:=(j-1)*ndel+i; r:=buf2[i];
  IF(spvar=0)THEN y[k]:=r; e[k]:=SQRT(AMAX1(r,1));
  END;
 Inc(j); nb:=nb+res; IF(nb>=nspch)THEN BREAK;
 END;
IF(spvar=0)THEN
 BEGIN a1:=1; a2:=MAX0(nb,a1);
 END;
y[1]:=0; e[1]:=1; CloseFile(fb); ok:=true; errset:=true;
END;
PROCEDURE TSpectrum.Read4Binary(VAR ok:BOOLEAN);
{Purpose: To read 4 bytes per word binary files, Version: Mar 30 2000}
VAR   res,ndel,i,j,k,jj,nb:INTEGER; fb:FILE OF INTEGER; r:REAL;
BEGIN ndel:=128; ok:=false; s:=sfn; IF(spvar=1)THEN s:=efn; nb:=0;
AssignFile(fb,s); {$I-} Reset(fb); {$I+};
IF(IOresult<>0)THEN
 BEGIN ShowMess(s+': opening failure; '+crc+
 'file may be Read_Only'); Exit;
 END;
IF(skiph)THEN
 BEGIN BlockRead(fb,buf4,ndel,res); nb:=nb+res;
 BlockRead(fb,buf4,ndel,res); nb:=nb+res;
 END;
j:=1; WHILE(NOT Eof(fb))DO
 BEGIN BlockRead(fb,buf4,ndel,res); FOR i:=1 TO res DO
  BEGIN k:=(j-1)*ndel+i; r:=buf4[i]; IF(revby)THEN
   BEGIN jj:=Round(r); ReverseBytes(jj,jj); r:=jj;
   END;
  IF(spvar=0)THEN y[k]:=r; e[k]:=SQRT(AMAX1(r,1));
  END;
 Inc(j); nb:=nb+res; IF(nb>=nspch)THEN BREAK;
 END;
IF(spvar=0)THEN
 BEGIN a1:=1; a2:=MAX0(nb,a1);
 END;
y[1]:=0; e[1]:=1; CloseFile(fb); ok:=true; errset:=true;
END;
PROCEDURE TSpectrum.ReadTxt(k1,k2:INTEGER; VAR ok:BOOLEAN);
{Reads into real array Ar from k1 to k2; if endfile / then kkk items}
VAR   i,j,L:INTEGER; ss,s:STRING; r:REAL; fv:TEXTFILE;
LABEL Lend;
BEGIN s:=sfn; IF(spvar=1)THEN s:=efn; AssignFile(fv,s); Reset(fv);
i:=k1; WHILE(NOT EoF(fv))DO
 BEGIN Readln(fv,s); L:=Length(s);
 IF(L=0)THEN CONTINUE ELSE IF(s[1]='*')THEN CONTINUE ELSE; s:=s+' ';
 FOR j:=1 TO Length(s) DO IF(s[j]=',')THEN s[j]:=' ';
 REPEAT FCompress(s); IF(Length(s)<>0)THEN
  BEGIN j:=Pos(' ',s); IF(j<>0)THEN
   BEGIN L:=j-1; ss:=Copy(s,1,L); Delete(s,1,L+1); r:=0;
   IF ss[1]='.' THEN ss:='0'+ss; Valz(ss,r,j); IF(spvar=0)THEN y[i]:=r;
   e[i]:=SQRT(AMAX1(r,0));
   IF(spvar=1)THEN
    CASE vartype OF
    0: e[i]:=r;
    1: e[i]:=SQRT(AMAX1(r,0));
    2: IF(r>0)THEN e[i]:=SQRT(1/r) ELSE e[i]:=0;
    ELSE END;
   Inc(i);
   END;
  {IF(i>k2)THEN GOTO Lend;}
  END;
 UNTIL Length(s)=0;
 IF(EoF(fv))THEN GOTO Lend;
 END;
Lend: ok:=true; CloseFile(fv); errset:=false;
IF(spvar=0)THEN
 BEGIN a1:=k1; a2:=k1; IF((i-1)>=k1)THEN a2:=i-1 ELSE ok:=false;
 END;
END;
PROCEDURE TSpectrum.ReadTxt3(k1,k2:INTEGER; VAR ok:BOOLEAN);
{Reads trees: x,y,err}
VAR   s:STRING; k,nn:INTEGER; fv:TEXTFILE;
LABEL Lend,Fail;
BEGIN AssignFile(fv,sfn); Reset(fv);  ok:=false;
FOR k:=1 TO 3 DO onl[k]:=0; errset:=false;
EnterStr(fv); s:=sedz;
fffins(nn,s); IF((nn>3)OR(nn<2))THEN
 BEGIN ShowMess('Warning, no channel-counts spectrum'); GOTO Fail;
 END;
k:=k1; IF(NOT chan1)THEN k:=Round(onl[1]);
Reset(fv); WHILE(NOT EoF(fv))DO
 BEGIN EnterStr(fv); s:=sedz; s:=s+' '; fffins(nn,s);
 IF(nn=0)THEN CONTINUE; IF(nn>1)THEN
  BEGIN x[k]:=onl[1]; y[k]:=onl[2]; e[k]:=SQRT(ABS(onl[2])+1);
  END;
 e[k]:=ABS(onl[3]);
 IF(nn>2)THEN
  BEGIN errset:=true;
  CASE vartype OF
  0: e[k]:=onl[3];
  1: e[k]:=SQRT(ABS(onl[3]));
  2: IF(onl[3]>0)THEN e[k]:=SQRT(1/onl[3]) ELSE e[k]:=0;
  ELSE END;
  END;
 Inc(k); IF(EoF(fv))THEN GOTO Lend;
 {IF((k>k2)OR(EoF(fv)))THEN GOTO Lend;}
 END;
Lend: ok:=true;
a1:=k1; a2:=k1; IF((k-1)>=k1)THEN a2:=k-1 ELSE ok:=false;
x1:=x[k1]; x2:=x[a2]; delx:=(x2-x1)/AMAX1(FLOAT(a2-a1),1);
celx1:=(a2-a1)/AMAX1((x2-x1),1); celx0:=a2-celx1*x2;
Fail: CloseFile(fv);
END;
PROCEDURE TSpectrum.fismax1;
{***  Purpose: To find the maxmin of the spectrum Version: Jan 6 2000}
VAR i:INTEGER; sx,sn,r:REAL;
BEGIN sx:=-1E14; sn:=1E14;
FOR i:=a1 TO a2 DO
 BEGIN r:=y[i]; IF(r > sx)THEN sx:=r;
 IF(r < sn)THEN sn:=r;
 END;
sn:=AMIN1(sn,0); spn:=sn; spx:=sx;
END;
FUNCTION TSpectrum.NoSpectrum(ns:INTEGER):BOOLEAN;
VAR ok:BOOLEAN; s:STRING;
BEGIN s:=sfn; UpCaseStr(s);
IF((spin)OR(s='DUMMY'))THEN ok:=false ELSE
 BEGIN ok:=true; ShowMess('No spectrum '+IntToStr(ns));
 END;
NoSpectrum:=ok;
END;
FUNCTION TSpectrum.chtod(ch:REAL):REAL;
{Purpose: To turn channel to d; Version: July 19 1994}
{ lambda = 2*d*sin(Theta) or d=t/(505*sin(theta)*F))}
VAR th:REAL;
BEGIN r:=chw; IF(r=0)THEN r:=1;
CASE tofang OF
0: Result:=(chw*(ch+delay))/(505.556*SIN(theta/radian)*FPath); {Dk}
1: BEGIN th:=(x[Round(ch)]+delay)*0.5;                         {angles}
   Result:=lambda*0.5/SIN((AMAX1(th,0.00001))/radian);
   END;
{1: Result:=lambda*0.5/SIN(((ch-1)*r+delay)/radian);}
   {d=chw0(k)*EXP(AMIN1(35,chw1(k)*ch))/constCl}
ELSE Result:=ch; END;
Result:=AMAX1(Result,0.00001);
END;
FUNCTION TSpectrum.dtoch(d:REAL):REAL;
{Purpose: To turn d to channel, Version: July 13 1994, Oct 3 1997}
{ q=inverse quadratic form: ((a*)*h)^2+...)     d=1/SQRT(q)}
VAR i:INTEGER; th:REAL;
BEGIN r:=chw; IF(r=0)THEN
 BEGIN ShowMess('Wrong channel width'); r:=1;
 END;
CASE tofang OF
0: Result:=d*505.556*SIN(theta/radian)*FPath/chw-delay;   {Dk}
1: BEGIN th:=radian*arcsin(AMIN1(lambda*0.5/d,1));  {angle}
   th:=th*2-delay; FOR i:=a11 TO a22-1 DO IF(rinfun(th,x[i],x[i+1]))THEN
    BEGIN Result:=i+(th-x[i])/(x[i+1]-x[i]);
    Exit;
    END;
   FOR i:=a1 TO a2-1 DO IF(rinfun(th,x[i],x[i+1]))THEN
    BEGIN Result:=i+(th-x[i])/(x[i+1]-x[i]); Exit;
    END;
   END;
ELSE Result:=AMAX1(d,1); END;
 {Result:=(radian*arcsin(AMIN1(lambda*q*0.5,1))-delay)/r+1;}
 {t=constCl/q12;clat=LN(t/chw0(k))/chw1(k)}
END;
PROCEDURE TSpectrum.Write2Binary(i1,i2:INTEGER; s:STRING);
{Purpose: To write 2 bytes per word binary files, Version: May 21 2001}
VAR res,ndel,i,j,k,kr,l:INTEGER; fb:FILE OF WORD;
BEGIN ndel:=256; AssignFile(fb,s);
{$I-} ReWrite(fb); {$I+}; ok:=(IOresult=0);
IF NOT ok THEN
 BEGIN ShowMess('Can not write to '+s); ok:=false; Exit;
 END;
j:=i2-i1+1; k:=j DIV ndel; kr:=j-ndel*k; IF(kr<>0)THEN Inc(k);
FOR j:=1 TO k DO
 BEGIN FOR i:=1 TO ndel DO
  IF(spvar=0)THEN buf2[i]:=Round(y[(j-1)*ndel+i])
             ELSE buf2[i]:=Round(SQR(e[(j-1)*ndel+i]));
 l:=ndel; IF((j=k)AND(kr<>0))THEN l:=kr; BlockWrite(fb,buf2,l,res);
 END;
CloseFile(fb); ok:=true;
END;
PROCEDURE TSpectrum.Write4Binary(i1,i2:INTEGER; s:STRING);
{Purpose: To write 4 bytes per word binary files, Version: May 21 2001}
VAR   res,ndel,i,j,k,kr,l:INTEGER; fb:FILE OF INTEGER;
BEGIN ndel:=128; ok:=false; AssignFile(fb,s);
{$I-} ReWrite(fb); {$I+}; ok:=(IOresult=0);
IF NOT ok THEN
 BEGIN ShowMess('Can not write to '+s); ok:=false; Exit;
 END;
j:=i2-i1+1; k:=j DIV ndel; kr:=j-ndel*k; IF(kr<>0)THEN Inc(k);
FOR j:=1 TO k DO
 BEGIN FOR i:=1 TO ndel DO
  IF(spvar=0)THEN buf4[i]:=Round(y[(j-1)*ndel+i])
             ELSE buf4[i]:=Round(SQR(e[(j-1)*ndel+i]));
 l:=ndel; IF((j=k)AND(kr<>0))THEN l:=kr; BlockWrite(fb,buf4,l,res);
 END;
CloseFile(fb); ok:=true;
END;
PROCEDURE TSpectrum.WriteTxt(i1,i2:INTEGER; s:STRING; nd:INTEGER;
          sc:BOOLEAN);
{Purpose: To write as text files, Version: May 21 2001}
VAR   i:INTEGER; fb:TEXTFILE; st:STRING; ok1:BOOLEAN; r:REAL;
BEGIN ok:=false; AssignFile(fb,s);
{$I-} ReWrite(fb); {$I+}; ok:=(IOresult=0); IF NOT ok THEN
 BEGIN ShowMess('Can not write to '+s); ok:=false; Exit;
 END;
s:=''; FOR i:=i1 TO i2 DO
 BEGIN IF(spvar=0)THEN r:=y[i] ELSE r:=SQR(e[i]); ok1:=false;
 IF(NOT sc)THEN
  BEGIN ConvertStr(r,25,nd); st:=sedz; Compress(st);
  END      ELSE ftos(r,MIN0(23,nd+7),nd,st);
 IF((Length(s)+Length(st)+1)>72)THEN
  BEGIN Writeln(fb,s); s:=''; ok1:=true;
  END;
 s:=s+st+' ';
 END;
IF(NOT ok1)THEN Writeln(fb,s); CloseFile(fb); ok:=true;
END;
PROCEDURE TSpectrum.WriteTxt3(i1,i2:INTEGER; ss:STRING; nd:INTEGER;
            sc,chang:BOOLEAN);
{Write as text file of strings: X, Y(counts), E(rror),
 Version: Nov 26 2001}
VAR i,j,k,n:INTEGER; fb:TEXTFILE; r:REAL; st,s:STRING;
BEGIN ok:=false; AssignFile(fb,ss);
{$I-} ReWrite(fb); {$I+}; ok:=(IOresult=0);
IF NOT ok THEN
 BEGIN ShowMess('Can not write to '+ss); ok:=false; Exit;
 END;
r:=ABS(spx)+1; r:=LN(r)/2.3+3; k:=Round(r)+nd;
n:=0; IF(chang)THEN n:=MAX0(nd,4);
IF(nd<>0)THEN Inc(k); FOR i:=i1 TO i2 DO
 BEGIN r:=i; IF(chang)THEN r:=x[i];
 IF(NOT sc)THEN Writeln(fb,r:8:n,' ',y[i]:k:nd,' ',e[i]:k:nd)
           ELSE
  BEGIN j:=MIN0(23,nd+7); ftos(r,j,nd,s); ftos(y[i],j,nd,st);
  s:=s+' '+st; ftos(e[i],j,nd,st); s:=s+' '+st; Writeln(fb,s);
  END;
 END;
CloseFile(fb); ok:=true;
END;
PROCEDURE TSpectrum.ReverseBytes(i:INTEGER; VAR o:INTEGER);
VAR ba:ARRAY[1..4] OF BYTE ABSOLUTE j; ka:ARRAY[1..4] OF BYTE;
    k:INTEGER;
BEGIN j:=i; ka[1]:=ba[3]; ka[2]:=ba[4]; ka[3]:=ba[1]; ka[4]:=ba[2];
FOR k:=1 TO 4 DO ba[k]:=ka[k]; o:=j;
END;
PROCEDURE TSpectrum.Vancal(np:INTEGER; tau,Fp,zd:REAL);
{Purpose: To adjust the effective standard vanadium spectrum efsp0
          The range between 0.73 and 22.0 Angstroem
 Version: 5.4.1989; Delphi=May 8 2000
 Fp = Time of Flight Path (in m)
 tau   = Channel Width (in mksec)
 zd    = Analyzer Start}
CONST p1:REAL=0.54124E+06;
      p2:REAL=0.3253446E+01;
      q1:REAL=0.2755642E+09;
      q2:REAL=0.585E+01;
      e:REAL =0.6602738E+01;
      tau0:REAL=128;
      z0:REAL=-8.5;
      efsp0:ARRAY[1..211] OF REAL = (
   143,  163,  176,  194,  210,  222,  259,  290,  335,  395,
   465,  527,  634,  770,  873, 1020, 1217, 1403, 1621, 1839,
  2120, 2391, 2697, 2921, 3239, 3728, 4151, 4526, 5102, 5689,
  6340, 7078, 8015, 8767, 9933,10960,12146,13427,14793,15993,
 17115,18402,19586,20839,21971,22915,23947,24928,25690,26291,
 26829,27423,27985,27822,28276,28439,28604,28473,28671,28965,
 28786,28747,28320,27874,27493,27231,26736,26296,25983,25669,
 25084,24792,24473,23838,23616,23178,22909,22396,22028,21597,
 21508,21067,21087,20477,20352,20145,19922,19660,19801,19588,
 19575,19278,19119,19059,19090,18587,18384,17947,17841,17766,
 17416,17121,17036,16760,16525,16334,16027,15894,15944,15808,
 15641,15479,15338,15079,15069,14619,14449,14329,14035,13647,
 13631,13203,13171,12913,12743,12475,12335,12091,11813,11544,
 11440,11127,10895,10742,10521,10374,10106, 9850, 9726, 9523,
  9302, 9160, 9018, 8795, 8746, 8611, 8527, 8298, 8158, 8020,
  7904, 7711, 7631, 7433, 7415, 7271, 7181, 6999, 6793, 6628,
  6485, 6397, 6153, 6074, 5922, 5823, 5710, 5672, 5700, 5708,
  5677, 5552, 5468, 5360, 5355, 5350, 5143, 5017, 4957, 4901,
  4835, 4732, 4711, 4604, 4505, 4436, 4311, 4263, 4183, 4147,
  4056, 3937, 3873, 3813, 3722, 3638, 3612, 3502, 3473, 3530,
  3557, 3533, 3435, 3490, 3424, 3373, 3352, 3266, 3226, 3186,
  3146);
VAR   i,j,k,n1,ic1,ic2,n0:INTEGER; efsp:ARRAY[1..1024] OF REAL;
      r,xn,q,znd,di,z1,z2,a,w1,w2,b,c,d,r1,rc:REAL;
LABEL L3,L4,L5,L6,L7;
BEGIN {form the interval bounds} r:=tau0/tau; xn:=r*(45+z0)-zd;
IF(xn<=0)THEN xn:=1; ic1:=Trunc(xn+0.5); xn:=r*(1094+z0)-zd;
IF(xn<0)THEN xn:=0; ic2:=Trunc(xn); IF(ic2>np)THEN ic2:=np;
IF(ic1<>1)THEN
 BEGIN n1:=ic1-1; FOR j:=1 TO n1 DO efsp[j]:=1;
 END;
r:=1/r; q:=252.7578*Fp;
FOR j:=ic1 TO ic2 DO
 BEGIN znd:=j+zd; xn:=znd*r-z0; IF(xn>=250)THEN GOTO L4;
 n0:=Trunc(xn); i:=n0-43; di:=xn-n0; IF(di<>0)THEN GOTO L5;
 efsp[j]:=efsp0[i]; GOTO L3;
 L5: z1:=efsp0[i]+efsp0[i+1]; z2:=efsp0[i-1]+efsp0[i+2];
 a:=(9*z1-z2)/16; IF(di<>0.5)THEN GOTO L6; efsp[j]:=a; GOTO L3;
 L6: w1:=efsp0[i+1]-efsp0[i]; w2:=efsp0[i+2]-efsp0[i-1];
 b:=(27*w1-w2)/24; c:=(z2-z1)/4; d:=(w2-3*w1)/6; xn:=di-0.5;
 efsp[j]:=((d*xn+c)*xn+b)*xn+a; GOTO L3;
 L4: r1:=znd*tau/q; IF(r1>=12)THEN GOTO L7;
 efsp[j]:=p1/powerRR(r1,p2); IF(r1<=10)THEN GOTO L3;
 efsp[j]:=efsp[j]*EXP(-e*SQR(LN(0.1*r1))); GOTO L3;
 L7: efsp[j]:=q1/powerRR(r1,q2); L3:;
 END;
a:=0; b:=1E10; FOR i:=1 TO ic2 DO
 BEGIN rc:=i*r-zd; k:=Round(rc); IF(infun(k,1,ic2))THEN
  BEGIN y[i]:=efsp[i]; a:=AMAX1(a,efsp[i]); b:=AMIN1(b,efsp[i]);
  END;
 END;
a1:=1; a2:=ic2; spx:=a; spn:=b; delay:=zd; FPath:=Fp; chw:=tau;
END;
PROCEDURE TSpectrum.polcc(n1,n11,n12,n2:INTEGER; VAR ae,be,x0:REAL);
{fit the linear background on the edges (var=const)  ae + be * (x - x0)
 Version: 2.11.1989; Apr 3 2000}
VAR i:INTEGER; wh,dis,disx,disy,disxx,disyx,z:REAL;
BEGIN wh:=1; ae:=0; be:=0; dis:=0; disx:=0; disy:=0; disyx:=0; disxx:=0;
FOR i:=n1 TO n11 DO
 BEGIN z:=i; dis:=dis+wh; disx:=disx+wh*z;
 END;
FOR i:=n12 TO n2 DO
 BEGIN z:=i; dis:=dis+wh; disx:=disx+wh*z;
 END;
x0:=disx/dis; FOR i:=n1 TO n11 DO
 BEGIN z:=i-x0; disxx:=disxx+wh*SQR(z);
 disy:=disy+wh*y[i]; disyx:=disyx+wh*y[i]*z;
 END;
FOR i:=n12 TO n2 DO
 BEGIN z:=i-x0; disxx:=disxx+wh*SQR(z);
 disy:=disy+wh*y[i]; disyx:=disyx+wh*z*y[i];
 END;
ae:=disy/dis; be:=disyx/disxx;
END;
PROCEDURE TSpectrum.corbc(l,ll:INTEGER; VAR nbl,nz:INTEGER);
{deletes blunders, if there are any;   version 85.10.09}
VAR   lmod2,lsz:BOOLEAN; i,j,k,m:INTEGER;
      wh,r1,r2,r3,r4,r5,sj,err,a,b,chim,value:REAL;
LABEL L91,L83;
 PROCEDURE whdec(i:INTEGER);
 {calculates the weight of i-th point of spectrum, version 85.10.30}
 BEGIN wh:=ABS(e[i]); IF(wh<>0)THEN wh:=1/(wh*wh);
 END;
{---------------------------------------------------------------}
BEGIN nz:=0; nbl:=0; m:=ll-l; IF(m<2)THEN Exit;
{delete the blunders} FOR i:=2 TO m DO
 BEGIN j:=l+i-1; r1:=y[j-1]+y[j+1]; r3:=y[j-1]*y[j+1];
 r4:=y[j-1]-y[j+1]; IF(r3=0)THEN r1:=2*ABS(r4);
 sj:=y[j]; IF((sj<>0)AND(r1<smast))THEN CONTINUE;
 r2:=ABS(r4)/(ABS(r1)+1); IF(r3=0)THEN r2:=0;
 lsz:=(r2<0.25)AND((sj>r1)OR(sj<(r1*0.25))); lsz:=lsz OR(sj=0);
 lmod2:=(i=2)OR(i=m)OR lsz; IF(lmod2)THEN GOTO L83;
 {compute the value of chebyshev polynomial which fits sp(j),j:=+-2,1}
 err:=2*SQRT(ABS(r1));
 IF((y[j-1]-y[j-2])<=err)THEN GOTO L91;
 IF((y[j]-y[j-1])<=err)THEN GOTO L91;
 IF((y[j]-y[j+1])<=err)THEN GOTO L91;
 IF((y[j+1]-y[j+2])<=err)THEN GOTO L91; CONTINUE;
 L91: r2:=r1+y[j-2]; r5:=r4-y[j-2]*2; a:=(3*r2+r5)/7; b:=(3*a-r2)*0.5;
 {compute chi-square for the points j+-2,j+-1}
 whdec(j-2); r1:=SQR(a-2*b-y[j-2])*wh;
 whdec(j+1); r1:=r1+SQR(a+b-y[j+1])*wh;
 whdec(j-1); r1:=r1+SQR(a-b-y[j-1])*wh;
 IF(r1>chibc)THEN CONTINUE;
 L83: IF(lmod2)THEN a:=r1*0.5; whdec(j); r1:=a-y[j]; r2:=(r1*wh)*r1;
 chim:=r2; value:=AMAX1(a,0); k:=j; IF(chim<=order)THEN CONTINUE;
 IF(sj=0)THEN Inc(nz); y[k]:=value; Inc(nbl);
 END;
END;
PROCEDURE TSpectrum.CalculateBack(VAR j1,j2,l:INTEGER);
{accounts for background, version Dec 10 2002}
VAR amx,er,ur:REAL; m,ll,lr,i,j:INTEGER;
BEGIN amx:=-1E10; FOR i:=j1 TO j2 DO amx:=AMAX1(y[i],amx);
IF(amx<=smast*3)THEN
 BEGIN ShowMess(' Too low statistics for model '+
  IntToStr(Round(amx))); Exit;
 END;
m:=j2-j1+1; FOR i:=j1 TO j2 DO onl[i-j1+1]:=(1000/amx)*y[i];
ll:=1; ur:=(ABS(onl[1])+ABS(onl[2]))*0.5; er:=SQRT(ur)*3;
FOR i:=1 TO (m DIV 2) DO
 BEGIN j:=i+2; IF(ABS(ABS(onl[j])-ur)>er)THEN BREAK; Inc(ll);
 END;
lr:=1; ur:=(ABS(onl[m])+ABS(onl[m-1]))*0.5; er:=SQRT(ur)*3;
m:=j2-j1+1; FOR i:=1 TO (m DIV 2) DO
 BEGIN j:=m-i-1; IF(ABS(ABS(onl[j])-ur)>er)THEN BREAK; Inc(lr);
 END;
{ShowMess(IntToStr(ll)+'  '+IntToStr(lr));}
IF(ll>5)THEN
 BEGIN j1:=j1+ll-5; ll:=5;
 END;
IF(lr>5)THEN
 BEGIN j2:=j2-lr+5; lr:=5;
 END;
l:=MIN0(ll,lr);
END;
PROCEDURE TSpectrum.CheckForSingleZeroes;
VAR i,nt,ns:INTEGER;
BEGIN nt:=0; ns:=0; FOR i:=a11+1 TO a22-1 DO
 BEGIN IF(y[i]=0)THEN Inc(nt);
 IF((y[i]=0)AND(y[i-1]<>0)AND(y[i+1]<>0))THEN Inc(ns);
 END;
IF(ns<>0)THEN
 BEGIN Writeln(fout,'Total number of zeroes= ',nt:4,
 ', single zeroes= ',ns:4);
 ShowMess('There are single zeroes in spectrum= '+IntToStr(ns));
 END;
END;
PROCEDURE TSpectrum.CheckErrorsFor0;
VAR i:INTEGER; sum:REAL; s:STRING;
BEGIN sum:=0; FOR i:=a1 TO a2 DO sum:=sum+e[i];
IF(sum=0)THEN errset:=false;
IF(NOT errset)THEN
 BEGIN s:='Errors=0. Make Poisson variance?';
 i:=MessageDlg(s,MtWarning,[mbYes,mbNo],0); IF(i=mrYes)THEN
  FOR i:=a1 TO a2 DO e[i]:=SQRT(ABS(y[i]+1));
 END;
END;
{----------}
BEGIN chibc:=1.25; order:=10; smast:=20; covar:=0.01;
skiph:=false; revby:=false; chan1:=true;
END.
